from pyspark.sql import SparkSession
from pyspark.sql.functions import split, explode, col, udf, lit
from pyspark.sql.types import StructType, StructField, StringType, IntegerType, BinaryType, LongType, BooleanType, TimestampType, DoubleType
from pyspark.sql import DataFrame as SparkDataFrame
import pandas as pd

# ensure consistency with transformations.add_metadata
meta_cols = [ ',\n"META_file_name" VARCHAR','"META_partition_date" TIMESTAMP','"META_processing_date_utc" TIMESTAMP','"META_loading_date_utc" TIMESTAMP default sysdate']

spark_meta_cols = {
    'META_file_name': StringType(),
    'META_partition_date': TimestampType(),
    'META_processing_date_utc': TimestampType()
}

spark_to_rs_meta_cols = [',\n"META_loading_date_utc" TIMESTAMP default sysdate']

mapping_table_sqlserver_redshift = {
    'bigint': 'BIGINT',
    'bigint IDENTITY':'BIGINT',
    'bit':'BOOLEAN',
    'datetime':'TIMESTAMP',
    'int':'BIGINT',
    'numeric':'NUMERIC',
    'nvarchar': 'VARCHAR',
    # in cogenius, version is captured as timestamp which we arbitrarily translate to varchar
    'timestamp': 'VARCHAR(256)',
    'varchar': 'VARCHAR'
}

mapping_table_pandas_redshift = {
    'int64': 'BIGINT',
    'float64': 'DOUBLE PRECISION',
    'object': 'VARCHAR(MAX)'
}

mapping_table_pandas_pyspark = {
    'int64': IntegerType(),
    'float64': DoubleType(),
    'object': StringType(),
    'bool': BooleanType(),
    'datetime64': TimestampType() 
}


mapping_pyspark_to_redshift = {
    'integer': 'BIGINT',
    'string': 'VARCHAR(MAX)',
    'double': 'DOUBLE PRECISION',
    'timestamp': 'TIMESTAMP',
    'boolean': 'BOOLEAN'
}

def csv_to_flyway_ddl(cols, dtypes, target_table, flyway_schema='ingest_schema'):
    """
        ARGUMENTS
            cols: df.columns
            dtypes: df.dtypes
        RETURNS 
            flyway ddl 
    """
    target_cols = []
    d = dict(zip(cols, dtypes))
    for key, val in d.items():
        target_type = mapping_table_pandas_redshift[val.name]
        line = f'"{key}" {target_type}'
        target_cols.append(line)

    joined_cols = ",\n".join(target_cols)
    joined_meta_cols = ",\n".join(meta_cols)
    final_cols = joined_cols+joined_meta_cols

    flyway_ddl = f"""CREATE TABLE IF NOT EXISTS ${{{flyway_schema}}}.{target_table} (
        {final_cols}
    )

    DISTSTYLE AUTO;

    ALTER TABLE ${{{flyway_schema}}}.{target_table} owner to ${{username}};
    """
    return flyway_ddl

 ## function to convert cogenius sql server DDL to flyway compatible ddl
def ddl_sql_to_flyway_ddl(sql_ddl, target_table, flyway_schema='ingest_schema'):
    from simple_ddl_parser import DDLParser
    result = DDLParser(sql_ddl).run(output_mode='mssql')
    columns = result[0]['columns']
    target_cols = []
    for col in columns:
        source_name = col['name']
        source_type = col['type']
        source_size = col['size']
        target_type = mapping_table_sqlserver_redshift[source_type]
        if target_type in ['VARCHAR']:
            target_type=f'{target_type}({source_size})'
        if target_type in ['NUMERIC']:
            target_type=f'{target_type}{source_size}'
        line = f'"{source_name}" {target_type}'
        target_cols.append(line)

    joined_cols = ",\n".join(target_cols)
    joined_meta_cols = ",\n".join(meta_cols)
    final_cols = joined_cols+joined_meta_cols

    new_ddl = f"""CREATE TABLE IF NOT EXISTS ${{{flyway_schema}}}.{target_table} (
        {final_cols}
    )

    DISTSTYLE AUTO;

    ALTER TABLE ${{{flyway_schema}}}.{target_table} owner to ${{username}};
    """
    return new_ddl


def write_flyway_ddl_to_file(flyway_ddl, target_table, flyway_schema='ingest'):
    from os import listdir
    from os.path import isfile, join

    flyway_dir = (f'/Users/pieterdepetter/GitHub-VEB/terra-etl/flyway/modules/redshift/{flyway_schema}')
    files = [f for f in listdir(flyway_dir) if isfile(join(flyway_dir, f))]
    files.sort()
    version, seq = files[-1].split('__')[0].split('_')
    new_version = int(version.split('V')[-1])+1
    new_version = f'V{new_version:03}'
    new_name = f'{new_version}_1__create_{target_table}.sql'
    with open(f'{flyway_dir}/{new_name}', 'w') as file:
        file.write(flyway_ddl)
    file.close()

def df_to_pyspark(df:pd.DataFrame):
    cols = df.columns
    dtypes = df.dtypes
    target_cols = []
    d = dict(zip(cols, dtypes))
    for key, val in d.items():
        key = key.strip().replace(' ', '_').replace('(','').replace(')','').replace(',','').replace(';','').replace('{','').replace('}','').replace('\n','').replace('\t','').replace('=','')
        target_type = mapping_table_pandas_pyspark[val.name]
        line = StructField(key, target_type, True)
        target_cols.append(line)
    for key, val in spark_meta_cols.items():
        line = StructField(key, val, True)
        target_cols.append(line)
    pyspark_schema = StructType(target_cols)
    return pyspark_schema


def pyspark_to_flyway_ddl(pyspark_schema, target_table, flyway_schema='ingest_schema', ):
    json_sch = pyspark_schema.jsonValue()
    target_cols = []
    for field in json_sch['fields']:
        target_name = field['name']
        target_type = mapping_pyspark_to_redshift[field['type']]
        line = f'"{target_name}" {target_type}'
        target_cols.append(line)

    joined_cols = ",\n".join(target_cols)
    joined_meta_cols = ",\n".join(spark_to_rs_meta_cols)
    final_cols = joined_cols+joined_meta_cols

    new_ddl = f"""CREATE TABLE IF NOT EXISTS ${{{flyway_schema}}}.{target_table} (
        {final_cols}
    )

    DISTSTYLE AUTO;

    ALTER TABLE ${{{flyway_schema}}}.{target_table} owner to ${{username}};
    """
    return new_ddl







import boto3
from urllib.parse import unquote
import os, sys, time
import logging
from io import BytesIO
import pandas as pd

class s3Client():

    def __init__(self, s3_bucket:str) -> None:
        self.bucket_name = s3_bucket
        self.bucket = self._create_bucket_resource()
        self.client = self._create_client()

    def _create_client(self):
        logging.info('Creating boto3 s3 client')
        s3_client = boto3.client("s3")
        logging.info('Succesfully created s3 client')
        return s3_client

    def _create_bucket_resource(self):
        logging.info('Creating boto3 s3 bucket resource')
        s3 = boto3.resource('s3')
        bucket = s3.Bucket(self.bucket_name)
        logging.info('Succesfully created bucket resource')
        return bucket

    def list_files(self, s3_prefix:str, file_type:str, max_files=1000000000000) -> list:
        """
            ARGUMENTS
                s3_prefix: folder/subfolder/
                file_type: eg. csv, xlsx, parquet
                max_files: optional, unless max_files is specified, all files are returned. 

            RETURNS 
                List of files from s3_prefix, filtering by type. Files are fetched per chunks of 1000. 
            
            ADDITIONAL INFO 
            Files are automatically unquoted and returned as full path ( ie. s3:://bucket/s3_prefix/filename.extension)
        """
        
        kwargs = {
            "Bucket": self.bucket_name,
            "Delimiter": "|",
            "EncodingType": "url",
            "MaxKeys": 1000,
            "Prefix": s3_prefix,
            "FetchOwner": False
        }
        truncated = True
        files = []
        counter = 0
        while truncated and counter < max_files:
            counter += 1000
            response = self.client.list_objects_v2(**kwargs)
            truncated = response["IsTruncated"]
            if "Contents" in response:
                for record in response["Contents"]:
                    filename = record["Key"]
                    files.append(filename)
                if truncated:
                    kwargs["ContinuationToken"] = response["NextContinuationToken"]
        files = files[:max_files]
        files = [x for x in files if file_type in x]
        files = [unquote(x) for x in files]
        files = ['s3://'+self.bucket_name+'/'+x for x in files]
        return files

    def _delete_prefix(self, s3_prefix:str ) -> None:
        """
            Delete all objects with a certain prefix
        """
        logging.info(f'Starting delete operation for prefix {s3_prefix}')
        self.bucket.objects.filter(Prefix=s3_prefix).delete()
        logging.info(f'Succesfully delete all objects for prefix {s3_prefix}')
        return
    
    def _prefix_exist_and_not_empty(self, s3_prefix:str) -> bool:
        '''
            S3_prefix should exist. 
            S3_prefix should not be empty.

            Returns boolean
        '''
        if not s3_prefix.endswith('/'):
            s3_prefix = s3_prefix+'/' 
        resp = self.client.list_objects(Bucket=self.bucket_name, Prefix=s3_prefix, Delimiter='/', MaxKeys=1)
        return 'Contents' in resp

    def delete_prefix_if_exist(self, s3_prefix:str) -> None:
        """
            ARGUMENTS
                s3_prefix: folder/subfolder/

            RETURNS 
                None

            ADDITIONAL INFO
                Checks whether the prefix contains at least 1 file or folder. If not, prefix is considered non-existent.
                If yes, prefix ( ie all folders, subfolders and files) will be deleted. 
        """
        try: 
            folder_exists = self._prefix_exist_and_not_empty(s3_prefix)
            if folder_exists:
                self._delete_prefix(s3_prefix)
                logging.warning(f'Found and deleted prefix {s3_prefix}.')
            else:
                logging.warning(f'Partition {s3_prefix} does not exist and will be created')
        except Exception as e:
            logging.error('Issue checking if prefixs exists already and/or deleting. Exiting...')
            logging.error(e)
            sys.exit(1)
        return

    def upload_local_file(self, local_file:str, s3_prefix: str) -> None:
        """
            ARGUMENTS
                local_file: ./data/file.ext
                s3_prefix: target_folder/target_subfolder

            RETURNS
                None

            ADDITIONAL INFO
                Use this function to upload a single file from local to s3. The local filename will be retained. 
        """
        target_file_name = local_file.split('/')[-1]
        target_file = f'{s3_prefix}{target_file_name}'
        try: 
            logging.info(f'Uploading file ...')
            self.client.upload_file(local_file, self.bucket_name, target_file)
            logging.info(f'Uploaded {local_file} to {target_file}')
        except Exception as e: 
            logging.error(f'Issue uploading {local_file} to {target_file}')
            logging.error(e)
            sys.exit(1)
        return

    def df_to_parquet(self, df: pd.DataFrame, s3_prefix:str, s3_file_name ) -> None:
        """
            ARGUMENTS
                df: pandas dataframe
                s3_prefix: target_folder/target_subfolder
                s3_file_name: name of the file without extension. A timestamp will be added within the function. 

            RETURNS
                None

            ADDITIONAL INFO
                DF will be read into memory and stored in S3 as parquet under key s3_prefix/1562388.0020_s3_filename.parquet
        """
        timestamp = round(time.time(), 4)
        parquet_buffer = BytesIO()
        df.to_parquet(parquet_buffer, index=False, allow_truncated_timestamps=True)
        s3_key = f'{s3_prefix}{timestamp}_{s3_file_name}.parquet'
        parquet_buffer.seek(0)
        self.bucket.upload_fileobj(parquet_buffer, s3_key)
        return
        
        